/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.impl;

/**
 * Utility methods for radix manipulation.
 * 
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 */
public final class Radix {

    private static final int HEX_RADIX = 16;
    private static final int SHIFT = 4;
    private static final int MASK = 0xf;

    /**
     * Prevent instantiation of utility class.
     */
    private Radix() {

    }

    /**
     * Returns the value of a given digit in a given radix.
     *
     * @param digit the digit
     * @param radix the radix
     * @return the corresponding value
     */
    public static int digitToValue(final char digit, final int radix) {

        if (digit < 'a') {

            if (digit < '0' || digit > '9') { throw new IllegalArgumentException("illegal digit " + digit + " for radix " + radix); }
            return digit - '0';
        }
        if (digit > 'z') { throw new IllegalArgumentException("illegal digit " + digit + " for radix " + radix); }
        return digit - 'a' + '9' - '0' + 1;
    }

    /**
     * Returns the digit corresponding to a given value in a given radix.
     *
     * @param value the value
     * @param radix the radix
     * @return the corresponding digit
     */
    public static char valueToDigit(final int value, final int radix) {

        if (value < 0 || value >= radix) { throw new IllegalArgumentException("illegal digit " + value + " for radix " + radix); }

        if (value < '9' - '0' + 1) { return (char) ('0' + value); }
        return (char) ('a' + value - ('9' - '0' + 1));
    }

    /**
     * Returns the number of bits needed to represent a given number in binary.
     *
     * @param n the number
     * @return the number of bits needed to represent the number in binary
     */
    public static int bitsInBinaryRepresentation(final int n) {

        final double double_value = Integer.valueOf(n).doubleValue();

        return (int) Math.floor(logBase2(double_value)) + 1;
    }

    /**
     * Returns the number of bits needed to represent a digit using the given radix.
     *
     * @param radix the number
     * @return the number of bits needed to represent a digit
     */
    public static int bitsPerDigit(final int radix) {

        return bitsInBinaryRepresentation(radix - 1);
    }

    /**
     * Returns the logarithm of a given number to base 2.
     *
     * @param n the number
     * @return the log of the number to base 2
     */
    public static double logBase2(final double n) {

        return log(n, 2.0);
    }

    /**
     * Returns the logarithm of a given number to a given base.
     *
     * @param n the number
     * @param base the base
     * @return the log of the number to the given base
     */
    public static double log(final double n, final double base) {

        if (n <= 0) { throw new IllegalArgumentException(); }

        return Math.log(n) / Math.log(base);
    }

    /**
     * Returns a representation of a byte array as a Hex string.
     *
     * @param bytes the byte array
     * @return a string representation
     */
    public static String toHexString(final byte[] bytes) {

        final StringBuilder buffer = new StringBuilder();

        for (final byte element : bytes) {

            buffer.append(valueToDigit(element >>> SHIFT & MASK, HEX_RADIX)); // 4 most significant bits - shift and mask.
            buffer.append(valueToDigit(element & MASK, HEX_RADIX)); // 4 least significant bits - mask.
        }
        return buffer.toString();
    }
}
