/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.p2p.util;

import uk.ac.standrews.cs.nds.p2p.interfaces.IHash;

/**
 * Generator of random bytes using a hash function.
 * 
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class RandomGenerator {

    private static int count = 0;

    private boolean deterministic = false;
    private final IHash hash;

    /**
     * Initialises the generator with a given hash function.
     * @param hash a hash function
     */
    public RandomGenerator(final IHash hash) {

        this.hash = hash;
    }

    /**
     * Generates a random byte array.
     * 
     * @param number_of_bytes the number of bytes required
     * @return the random bytes
     */
    public byte[] randomBytes(final int number_of_bytes) {

        final int bytes_in_hash = hash.getHashSize();

        final byte[] bytes = new byte[number_of_bytes];

        int i = 0;
        for (; i < number_of_bytes / bytes_in_hash; i++) {

            setRandomBytes(bytes, i, bytes_in_hash);
        }

        // Fill in remaining bytes if necessary.
        final int bytes_left = number_of_bytes % bytes_in_hash;
        if (bytes_left > 0) {
            setRandomBytes(bytes, i, bytes_left);
        }

        return bytes;
    }

    private void setRandomBytes(final byte[] bytes, final int i, final int bytes_to_copy) {

        String seed;

        if (deterministic) {
            seed = "seed" + count++;
        }
        else {
            seed = randomString();
        }

        final byte[] hashed = hash.hash(seed.getBytes());
        System.arraycopy(hashed, 0, bytes, i * hashed.length, bytes_to_copy);
    }

    /**
     * Returns a random string.
     * @return a random string
     */
    public String randomString() {

        return String.valueOf(System.currentTimeMillis()) + String.valueOf(Runtime.getRuntime().freeMemory());
    }

    /**
     * Returns true if deterministic.
     * @return true if deterministic
     */
    public boolean isDeterministic() {

        return deterministic;
    }

    /**
     * Sets whether the generator should operate deterministically.
     * @param deterministic true if the generator should operate deterministically
     */
    public void setDeterministic(final boolean deterministic) {

        this.deterministic = deterministic;
    }
}
