/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.remote_management;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import uk.ac.standrews.cs.nds.util.ClassPath;

import com.mindbright.ssh2.SSH2Exception;

/**
 * Record describing a remote host and, optionally, SSH connection credentials and/or remote references to an application that is running or could run on it.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class HostDescriptor {

    private final String host;
    private int port;
    private SSH2ConnectionWrapper ssh_client_wrapper;
    private PlatformDescriptor platform_descriptor;
    private ClassPath class_path;
    private URL[] application_urls;
    private HostState host_state;
    private Object application_reference;
    private Process process;
    private final Map<String, String> scan_results;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the network address.
     * @return the network address
     */
    public String getHost() {

        return host;
    }

    /**
     * Returns the network port.
     * @return the network port, or zero if not set
     */
    public int getPort() {

        return port;
    }

    /**
     * Sets the network port.
     * @param port the network port
     */
    public void setPort(final int port) {

        this.port = port;
    }

    /**
     * Returns the SSH connection credentials.
     * @return the SSH connection credentials, or null if not set
     */
    public SSH2ConnectionWrapper getSSHClientWrapper() {

        return ssh_client_wrapper;
    }

    /**
     * Returns the descriptor for the remote platform.
     * @return the descriptor for the remote platform, or null if not set
     */
    public PlatformDescriptor getPlatformDescriptor() {

        return platform_descriptor;
    }

    /**
     * Returns a classpath for the remote host with which an application can be run. This is null if {@link #application_urls} is not null.
     * @return a classpath for the remote host with which an application can be run, or null if not set
     */
    public ClassPath getClassPath() {

        return class_path;
    }

    /**
     * Returns an array of URLs for libraries against which an application can be run. This is null if {@link #class_path} is not null.
     * @return an array of URLs for libraries against which an application can be run, or null if not set
     */
    public URL[] getApplicationUrls() {

        // Clone to avoid exposing mutable state.
        return application_urls != null ? application_urls.clone() : null;
    }

    /**
     * Returns the state of the remote host and the application. Of course this is only an educated guess.
     * @return the state of the remote host and the application
     */
    public HostState getHostState() {

        return host_state;
    }

    /**
     * Sets the state of the remote host and the application.
     * @param host_state the state of the remote host and the application
     */
    public void setHostState(final HostState host_state) {

        this.host_state = host_state;
    }

    /**
     * Returns a remote reference to the application running on the remote host.
     * @return a remote reference to the application running on the remote host, or null if not set
     */
    public Object getApplicationReference() {

        return application_reference;
    }

    /**
     * Sets a remote reference to the application running on the remote host.
     * @param application_reference a remote reference to the application running on the remote host
     */
    public void setApplicationReference(final Object application_reference) {

        this.application_reference = application_reference;
    }

    /**
     * Returns a handle on the application process running on the remote host.
     * @return a handle on the application process running on the remote host, or null if not set
     */
    public Process getProcess() {

        return process;
    }

    /**
     * Sets a handle on the application process running on the remote host.
     * @param process a handle on the application process running on the remote host
     */
    public void setProcess(final Process process) {

        this.process = process;
    }

    /**
     * Returns a map containing results recorded by application-specific scanners.
     * @return a map containing results recorded by application-specific scanners
     */
    public Map<String, String> getScanResults() {

        return scan_results;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a new descriptor with a given address.
     * 
     * @param address the remote address
     */
    public HostDescriptor(final String address) {

        host = address;
        host_state = HostState.UNKNOWN;
        scan_results = new HashMap<String, String>();
    }

    /**
     * Creates a new descriptor with a given set of SSH credentials and a specified port.
     * 
     * @param ssh_client_wrapper a set of SSH credentials
     * @param port the port on which the application will run
     * @throws SSH2Exception if the platform of the remote host cannot be established
     */
    public HostDescriptor(final SSH2ConnectionWrapper ssh_client_wrapper, final int port) throws SSH2Exception {

        this(ssh_client_wrapper.getHost().getCanonicalHostName());

        this.port = port;
        this.ssh_client_wrapper = ssh_client_wrapper;

        platform_descriptor = PlatformDescriptor.getPlatform(ssh_client_wrapper);
    }

    /**
     * Creates a new descriptor with a given set of SSH credentials, a specified port and a classpath for the application.
     * 
     * @param ssh_client_wrapper a set of SSH credentials
     * @param port the port on which the application will run
     * @param class_path a classpath for the application
     * @throws SSH2Exception if the platform of the remote host cannot be established
     */
    public HostDescriptor(final SSH2ConnectionWrapper ssh_client_wrapper, final int port, final ClassPath class_path) throws SSH2Exception {

        this(ssh_client_wrapper, port);

        this.class_path = class_path;
    }

    /**
     * Creates a new descriptor with a given set of SSH credentials, a specified port and a classpath for the application.
     * 
     * @param ssh_client_wrapper a set of SSH credentials
     * @param port the port on which the application will run
     * @param application_urls an array of URLs for libraries against which the application can be run
     * @throws SSH2Exception if the platform of the remote host cannot be established
     */
    public HostDescriptor(final SSH2ConnectionWrapper ssh_client_wrapper, final int port, final URL[] application_urls) throws SSH2Exception {

        this(ssh_client_wrapper, port);

        this.application_urls = application_urls;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        return "host: " + host + "\n" + "port: " + port + "\n";
    }
}
