/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.remote_management;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeoutException;

import uk.ac.standrews.cs.nds.util.ClassPath;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;

import com.mindbright.ssh2.SSH2Exception;

/**
 * Description of a recognised OS platform.
 * 
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public final class PlatformDescriptor {

    /**
     * Linux.
     */
    public static final String NAME_LINUX = "Linux";

    /**
     * OS X.
     */
    public static final String NAME_MAC = "Mac OS X";

    /**
     * Windows.
     */
    public static final String NAME_WINDOWS = "Windows";

    /**
     * Unknown.
     */
    public static final String NAME_UNKNOWN = "Unknown";

    // -------------------------------------------------------------------------------------------------------

    private static final String OUTPUT_CLASS_NAME = StringConstants.getString("OUTPUT_CLASS_NAME");
    private static final String JAVAC_MAIN_CLASS_NAME = StringConstants.getString("JAVAC_MAIN_CLASS_NAME");

    private static final String DEFAULT_WGET_PATH_LINUX = StringConstants.getString("DEFAULT_WGET_PATH_LINUX");
    private static final String DEFAULT_WGET_PATH_MAC = StringConstants.getString("DEFAULT_WGET_PATH_MAC");

    private static final String DEFAULT_TEMP_PATH_LINUX = StringConstants.getString("DEFAULT_TEMP_PATH_LINUX");
    private static final String DEFAULT_TEMP_PATH_MAC = StringConstants.getString("DEFAULT_TEMP_PATH_MAC");

    private static final String CLASS_PATH_SEPARATOR_LINUX = StringConstants.getString("CLASS_PATH_SEPARATOR_LINUX");
    private static final String CLASS_PATH_SEPARATOR_MAC = StringConstants.getString("CLASS_PATH_SEPARATOR_MAC");
    private static final String CLASS_PATH_SEPARATOR_WINDOWS = StringConstants.getString("CLASS_PATH_SEPARATOR_WINDOWS");

    private static final String FILE_SEPARATOR_LINUX = StringConstants.getString("FILE_SEPARATOR_LINUX");
    private static final String FILE_SEPARATOR_MAC = StringConstants.getString("FILE_SEPARATOR_MAC");
    private static final String FILE_SEPARATOR_WINDOWS = StringConstants.getString("FILE_SEPARATOR_WINDOWS");

    // -------------------------------------------------------------------------------------------------------

    private static final String JAVA_EXECUTABLE_NAME = "java";
    private static final String JAVAC_EXECUTABLE_NAME_WINDOWS = "javac.exe";
    private static final String JAVAC_EXECUTABLE_NAME = "javac";

    private static final String PROBE_COMMAND = "uname";

    private static final String PROBE_REPLY_LINUX = "Linux";
    private static final String PROBE_REPLY_MAC = "Darwin";

    private static final int SSH_TIMEOUT = 30000;

    private static PlatformDescriptor current_platform = null;
    private static Map<SSH2ConnectionWrapper, PlatformDescriptor> cache = new HashMap<SSH2ConnectionWrapper, PlatformDescriptor>();

    // -------------------------------------------------------------------------------------------------------

    // Attributes that can be read directly on the local platform.
    private final String name;
    private final String file_separator;
    private final String class_path_separator;
    private final String temp_path;

    // Attributes that have to be inferred from the platform type.
    private final String wget_path;
    private final String path_quote;

    // -------------------------------------------------------------------------------------------------------

    private PlatformDescriptor(final String name, final String file_separator, final String class_path_separator, final String temp_path, final String wget_path, final String path_quote) {

        this.name = name;
        this.file_separator = file_separator;
        this.class_path_separator = class_path_separator;
        this.temp_path = temp_path;

        this.wget_path = wget_path;
        this.path_quote = path_quote;
    }

    private PlatformDescriptor(final String name, final String file_separator, final String class_path_separator, final String temp_path) {

        this(name, file_separator, class_path_separator, temp_path, getWgetPath(name), getPathQuote(name));
    }

    private PlatformDescriptor(final String name) {

        this(name, getFileSeparator(name), getClassPathSeparator(name), getTempPath(name));
    }

    private PlatformDescriptor() {

        this(NAME_UNKNOWN);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the platform that this Java instance is running on.
     * 
     * @return the platform that this Java instance is running on
     */
    public static synchronized PlatformDescriptor getPlatform() {

        if (current_platform == null) {

            String platform_name = System.getProperty("os.name");

            if (platform_name.startsWith(NAME_WINDOWS)) {
                platform_name = NAME_WINDOWS;
            }

            if (platform_name.equals(NAME_MAC) || platform_name.equals(NAME_LINUX) || platform_name.equals(NAME_WINDOWS)) {

                final String file_separator = System.getProperty("file.separator");
                final String class_path_separator = System.getProperty("path.separator");
                final String temp_path = System.getProperty("java.io.tmpdir");

                current_platform = new PlatformDescriptor(platform_name, file_separator, class_path_separator, temp_path);
            }
            else {
                current_platform = new PlatformDescriptor();
            }
        }

        return current_platform;
    }

    /**
     * Returns the platform of a given remote host.
     * 
     * @param connection_wrapper a set of SSH credentials
     * @return the platform of the given remote host
     * @throws SSH2Exception if the platform cannot be determined
     */
    public static PlatformDescriptor getPlatform(final SSH2ConnectionWrapper connection_wrapper) throws SSH2Exception {

        if (cache.containsKey(connection_wrapper)) { return cache.get(connection_wrapper); }

        // Probe remote machine to discover OS.

        // Examples of expected replies:
        //   Linux
        //   Darwin

        final Semaphore sync = new Semaphore(0);
        final StringBuffer remote_reply_buffer = new StringBuffer();

        // Processor to read the output from the remote process and extract the platform name.
        final IStreamProcessor extract_platform = new IStreamProcessor() {

            @Override
            public boolean processByte(final int byte_value) {

                // Accumulate the output.
                remote_reply_buffer.append((char) byte_value);

                if (byte_value == '\n') {
                    sync.release();
                }

                return false;
            }
        };

        String platform_name = NAME_UNKNOWN;

        try {
            // Clone the SSH connection to avoid interference with other use of this one.
            final SSH2ConnectionWrapper cloned = connection_wrapper.clone();
            ProcessInvocation.runProcess(PROBE_COMMAND, cloned, System.out, System.err, extract_platform, null, SSH_TIMEOUT);

            // Wait for the remote process to complete.
            sync.acquire();

            final String remote_reply = remote_reply_buffer.toString();

            if (remote_reply.contains(PROBE_REPLY_MAC)) {
                platform_name = NAME_MAC;
            }
            if (remote_reply.contains(PROBE_REPLY_LINUX)) {
                platform_name = NAME_LINUX;
            }
        }
        catch (final TimeoutException e) {
            Diagnostic.trace(DiagnosticLevel.FULL, "exception determining remote platform: " + e.getMessage());
        }
        catch (final InterruptedException e) {
            Diagnostic.trace(DiagnosticLevel.FULL, "exception determining remote platform: " + e.getMessage());
        }
        catch (final IOException e) {
            Diagnostic.trace(DiagnosticLevel.FULL, "exception determining remote platform: " + e.getMessage());
        }

        final PlatformDescriptor result = new PlatformDescriptor(platform_name);

        cache.put(connection_wrapper, result);
        return result;
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public String toString() {

        return getName();
    }

    /**
     * Returns the platform's name.
     * 
     * @return the platform's name
     */
    public String getName() {

        return name;
    }

    /**
     * Returns the file separator character for this platform.
     * 
     * @return the file separator character for this platform
     */
    public String getFileSeparator() {

        return file_separator;
    }

    /**
     * Returns the class path separator character for this platform.
     * 
     * @return the class path separator character for this platform
     */
    public String getClassPathSeparator() {

        return class_path_separator;
    }

    /**
     * Returns the 'temp' directory path for this platform.
     * 
     * @return the 'temp' directory path for this platform
     */
    public String getTempPath() {

        return temp_path;
    }

    /**
     * Returns the 'wget' path for this platform.
     * 
     * @return the 'wget' path for this platform
     */
    public String getWgetPath() {

        return wget_path;
    }

    /**
     * Returns the character used to quote a class path for this platform. This may be the empty string.
     * 
     * @return the character used to quote a class path for this platform
     */
    public String getPathQuote() {

        return path_quote;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a shell command to invoke the Java interpreter with the specified class and arguments, using the class path of the currently executing program.
     *
     * @param clazz the class to be executed
     * @param args the command line arguments
     * @return a shell command to execute the class
     * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
     * @throws UnknownPlatformException if the local platform cannot be determined
     */
    public String getJavaCommand(final Class<?> clazz, final List<String> args) throws FileNotFoundException, UnknownPlatformException {

        // TODO doesn't work on Mac with spaces in class path elements.
        return getJavaCommand(clazz, args, ClassPath.getCurrentClassPath(), null);
    }

    /**
     * Creates a shell command to invoke the Java interpreter with the specified class and arguments on the specified platform, using the specified class path.
     *
     * @param clazz the class to be executed
     * @param args the command line arguments
     * @param class_path the class path
     * @param java_library_path the path to be used for the Java library path (for native libraries), or null if none required
     * @return a shell command to execute the class
     * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
     * @throws UnknownPlatformException if the local platform cannot be determined
     */
    public String getJavaCommand(final Class<?> clazz, final List<String> args, final ClassPath class_path, final File java_library_path) throws FileNotFoundException, UnknownPlatformException {

        return getJavaCommand(clazz, args, class_path, java_library_path, null);
    }

    /**
     * Creates a shell command to invoke the Java interpreter with the specified class and arguments on the specified platform, using the specified class path.
     *
     * @param clazz the class to be executed
     * @param args the command line arguments
     * @param class_path the class path
     * @param java_library_path the path to be used for the Java library path (for native libraries), or null if none required
     * @param java_bin_path the path of the directory containing the Java executable, or null for unspecified
     * @return a shell command to execute the class
     * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
     * @throws UnknownPlatformException if the local platform cannot be determined
     */
    public String getJavaCommand(final Class<?> clazz, final List<String> args, final ClassPath class_path, final File java_library_path, final File java_bin_path) throws FileNotFoundException, UnknownPlatformException {

        final String quote = getPathQuote(name);

        final StringBuilder buffer = new StringBuilder();

        buffer.append(quote);
        if (java_bin_path != null) {
            buffer.append(java_bin_path.getAbsolutePath() + file_separator);
        }
        buffer.append(JAVA_EXECUTABLE_NAME);
        buffer.append(quote);

        if (java_library_path != null) {
            buffer.append(" -Djava.library.path=");
            buffer.append(java_library_path.getAbsolutePath());
        }

        buffer.append(" -cp ");
        buffer.append(class_path.toString(this));

        buffer.append(" ");
        buffer.append(clazz.getName());

        if (args != null) {
            for (final String arg : args) {
                buffer.append(" ");
                buffer.append(arg);
            }
        }

        return buffer.toString();
    }

    /**
     * Creates a shell command to invoke the Java compiler with the specified parameters.
     *
     * @param compilation_class_path the class path to be compiled against
     * @param output_dir_path the path of the output directory
     * @param compiler_flags the compiler flags
     * @param file_paths the paths of the files to be compiled
     * @return a shell command to invoke the compiler
     * @throws FileNotFoundException if the compiler executable cannot be found locally
     */
    public String getJavaCompilerCommand(final ClassPath compilation_class_path, final File output_dir_path, final List<String> compiler_flags, final List<File> file_paths) throws FileNotFoundException {

        final ClassPath current_class_path = ClassPath.getCurrentClassPath();
        current_class_path.append(compilation_class_path);

        final StringBuilder buffer = new StringBuilder();

        final String path_quote = getPathQuote(name);

        buffer.append(path_quote);
        buffer.append(getJavaCompilerName(name));
        buffer.append(path_quote);

        buffer.append(" -cp ");
        buffer.append(current_class_path.toString());

        buffer.append(" -d ");
        buffer.append(path_quote);
        buffer.append(output_dir_path.getAbsolutePath());
        buffer.append(path_quote);
        buffer.append(" ");

        for (final String element : compiler_flags) {
            buffer.append(element);
            buffer.append(" ");
        }

        for (final File element : file_paths) {
            buffer.append(path_quote);
            buffer.append(element.getAbsolutePath());
            buffer.append(path_quote);
            buffer.append(" ");
        }

        return buffer.toString();
    }

    /**
     * Returns the name of the main Java compiler class.
     * 
     * @return the name of the main Java compiler class
     */
    public static String getJavaCompilerMainClassName() {

        return JAVAC_MAIN_CLASS_NAME;
    }

    /**
     * Returns the name of the Java compiler output class.
     * 
     * @return the name of the Java compiler output class
     */
    public static String getJavaCompilerOutputClassName() {

        return OUTPUT_CLASS_NAME;
    }

    // -------------------------------------------------------------------------------------------------------

    private static String getJavaCompilerName(final String platform_name) {

        if (platform_name.equals(NAME_MAC) || platform_name.equals(NAME_LINUX)) { return JAVAC_EXECUTABLE_NAME; }
        if (platform_name.equals(NAME_WINDOWS)) { return JAVAC_EXECUTABLE_NAME_WINDOWS; }
        return "";
    }

    private static String getFileSeparator(final String platform_name) {

        if (platform_name.equals(NAME_MAC)) { return FILE_SEPARATOR_MAC; }
        if (platform_name.equals(NAME_LINUX)) { return FILE_SEPARATOR_LINUX; }
        if (platform_name.equals(NAME_WINDOWS)) { return FILE_SEPARATOR_WINDOWS; }
        return "";
    }

    private static String getClassPathSeparator(final String platform_name) {

        if (platform_name.equals(NAME_MAC)) { return CLASS_PATH_SEPARATOR_MAC; }
        if (platform_name.equals(NAME_LINUX)) { return CLASS_PATH_SEPARATOR_LINUX; }
        if (platform_name.equals(NAME_WINDOWS)) { return CLASS_PATH_SEPARATOR_WINDOWS; }
        return "";
    }

    private static String getPathQuote(final String platform_name) {

        if (platform_name.equals(NAME_MAC) || platform_name.equals(NAME_LINUX)) { return ""; }
        if (platform_name.equals(NAME_WINDOWS)) { return "\""; }
        return "";
    }

    private static String getTempPath(final String platform_name) {

        if (platform_name.equals(NAME_MAC)) { return DEFAULT_TEMP_PATH_MAC; }
        if (platform_name.equals(NAME_LINUX)) { return DEFAULT_TEMP_PATH_LINUX; }
        return "";
    }

    private static String getWgetPath(final String platform_name) {

        if (platform_name.equals(NAME_MAC)) { return DEFAULT_WGET_PATH_MAC; }
        if (platform_name.equals(NAME_LINUX)) { return DEFAULT_WGET_PATH_LINUX; }
        return "";
    }
}
