/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.remote_management;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeoutException;

import uk.ac.standrews.cs.nds.util.ClassPath;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.IActionWithResult;
import uk.ac.standrews.cs.nds.util.Timeout;

import com.mindbright.ssh2.SSH2ConsoleRemote;
import com.mindbright.ssh2.SSH2Exception;

/**
 * Utility class for creating local and remote OS processes.
 *
 * @author graham
 */
public final class ProcessInvocation {

    /**
     * Prevent instantiation of utility class.
     */
    private ProcessInvocation() {

    }

    // -------------------------------------------------------------------------------------------------------

    private static final int KILL_TIMEOUT = 30000;
    private static final int DEFAULT_SSH_CONNECTION_TIMEOUT = 10000;
    private static final int NODE_INSTANTIATION_TIMEOUT = 30000; // The timeout for node instantiation, in ms.
    private static final String SHELL_PATH = "/bin/sh";

    // -------------------------------------------------------------------------------------------------------

    // Methods to run local shell commands.

    /**
     * Runs the specified shell command in a new OS process. Output is written to local System.out and System.err.
     *
     * @param command the command to be executed
     * @return a handle to the new OS process
     * @throws IOException if an error occurs when reading output from the process
     */
    public static Process runProcess(final String command) throws IOException {

        return runProcess(command, System.out, System.err);
    }

    /**
     * Runs the specified command in a new OS process.
     *
     * @param command the command to be executed
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @return a handle to the new OS process
     * @throws IOException if an error occurs when reading output from the process
     */
    public static Process runProcess(final String command, final OutputStream output_stream, final OutputStream error_stream) throws IOException {

        final Process process;

        if (!PlatformDescriptor.getPlatform().getName().equals(PlatformDescriptor.NAME_WINDOWS)) {

            // If not on Windows, run the command in a new OS shell.
            final String[] wrapped_command = new String[]{SHELL_PATH, "-c", command};
            process = Runtime.getRuntime().exec(wrapped_command);
        }
        else {

            process = Runtime.getRuntime().exec(command);
        }

        final PrintStream output_print_stream = new PrintStream(output_stream);
        final PrintStream error_print_stream = new PrintStream(error_stream);

        // Get streams reading from the standard out and standard error for the new process.
        final BufferedReader command_output_stream = new BufferedReader(new InputStreamReader(new BufferedInputStream(process.getInputStream())));
        final BufferedReader command_error_stream = new BufferedReader(new InputStreamReader(new BufferedInputStream(process.getErrorStream())));

        // Start a thread to capture any output and write it locally.
        makeReaderThread(command_output_stream, output_print_stream).start();

        // Start a thread to capture any errors and write them locally.
        makeReaderThread(command_error_stream, error_print_stream).start();

        return process;
    }

    // -------------------------------------------------------------------------------------------------------

    // Methods to run remote shell commands.

    /**
     * Runs the specified command on a remote machine accessed via SSH. Output from the remote process is passed to standard out and standard error.
     *
     * @param command the command to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @return a handle to the remote process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the command has not been invoked within the default timeout period
     */
    public static Process runProcess(final String command, final SSH2ConnectionWrapper connection_wrapper) throws SSH2Exception, IOException, TimeoutException {

        return runProcess(command, connection_wrapper, DEFAULT_SSH_CONNECTION_TIMEOUT);
    }

    /**
     * Runs the specified command on a remote machine accessed via SSH. Output from the remote process is passed to standard out and standard error.
     *
     * @param command the command to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param timeout the timeout for completing the issue of the command (NB not for the remote process to complete) in milliseconds
     *
     * @return a handle to the remote process
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the command has not been invoked within the specified timeout period
     */
    public static Process runProcess(final String command, final SSH2ConnectionWrapper connection_wrapper, final int timeout) throws SSH2Exception, IOException, TimeoutException {

        return runProcess(command, connection_wrapper, System.out, System.err, timeout);
    }

    /**
     * Runs the specified command on a remote machine accessed via SSH. Output from the remote process is passed to standard out and standard error.
     *
     * @param command the command to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @return a handle to the remote process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the command has not been invoked within the default timeout period
     */
    public static Process runProcess(final String command, final SSH2ConnectionWrapper connection_wrapper, final OutputStream output_stream, final OutputStream error_stream) throws SSH2Exception, IOException, TimeoutException {

        return runProcess(command, connection_wrapper, output_stream, error_stream, DEFAULT_SSH_CONNECTION_TIMEOUT);
    }

    /**
     * Runs the specified command on a remote machine accessed via SSH, using public key authentication. Output from the remote process is
     * passed through the specified processors, and then directed to the specified streams.
     *
     * Based on example code from MindTerm (http://www.mindterm.com/).
     *
     * @param command the command to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @param timeout the timeout for completing the issue of the command (NB not for the remote process to complete) in milliseconds
     *
     * @return a handle to the remote process
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the command has not been invoked within the specified timeout period
     */
    public static Process runProcess(final String command, final SSH2ConnectionWrapper connection_wrapper, final OutputStream output_stream, final OutputStream error_stream, final int timeout) throws SSH2Exception, IOException, TimeoutException {

        return runProcess(command, connection_wrapper, output_stream, error_stream, null, null, timeout);
    }

    /**
     * Runs the specified command on a remote machine accessed via SSH.
     *
     * @param command the command to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @param output_processor a processor for the standard output from the new process
     * @param error_processor a processor for the error output from the new process
     * @param timeout the timeout for completing the issue of the command (NB not for the remote process to complete) in milliseconds
     * @return a handle to the remote process
     * 
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws IOException if an error occurs when reading output from the process
     * @throws TimeoutException if the command has not been invoked within the specified timeout period
     */
    public static Process runProcess(final String command, final SSH2ConnectionWrapper connection_wrapper, final OutputStream output_stream, final OutputStream error_stream, final IStreamProcessor output_processor, final IStreamProcessor error_processor, final int timeout) throws SSH2Exception,
                    IOException, TimeoutException {

        /*
         * The MindTerm package supports the creation of a remote process, but not the later destruction of that process should it become necessary. So:
         *
         * 1. Prefix the remote user command with a shell command to echo a unique timestamp.
         * 2. Execute the combined commands.
         * 3. Store the timestamp so it can be used in a remote 'kill' command if the process needs to be destroyed.
         */

        Diagnostic.trace(DiagnosticLevel.RUN, "Remote command: ", command);

        // Get a timestamp - don't care about the time but it's an easy way to get a unique string.
        final String timestamp = String.valueOf(System.currentTimeMillis());

        // Prefix the command with code to echo the timestamp.
        // This will appear in a process listing and so can be used to identify the process if it needs to be destroyed..
        final String modified_command = makeLabelledCommand(command, timestamp);

        // Get streams that process the output and error streams.
        final OutputStream intercept_output_stream = makeInterceptStream(output_stream, output_processor);
        final OutputStream intercept_error_stream = makeInterceptStream(error_stream, error_processor);

        // Construct the action to run the labelled command in a new thread.
        final IActionWithResult run_command_action = makeRunCommandAction(connection_wrapper, modified_command, intercept_output_stream, intercept_error_stream);

        // Run the action subject to timeout.
        final Object result = new Timeout().performActionWithTimeout(run_command_action, timeout);

        // If the execution of the action resulted in an exception it will be returned as the result of the action.
        // This works round the limitations of the Timeout class which doesn't allow actions to throw exceptions.
        if (result instanceof SSH2Exception) { throw (SSH2Exception) result; }
        if (result instanceof IOException) { throw (IOException) result; }
        if (result instanceof RuntimeException) { throw (RuntimeException) result; }

        // Return a Java process handle that allows the remote process to be killed.
        return makeProcessForConsole(connection_wrapper, timestamp, (SSH2ConsoleRemote) result);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Attempts to kill all remote processes whose name includes the specified string.
     * 
     * @param match the string to match
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * 
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws IOException if an error occurs when communicating with the remote machine
     * @throws TimeoutException if the kill command has not been invoked within the default timeout period
     */
    public static void killMatchingProcesses(final String match, final SSH2ConnectionWrapper connection_wrapper) throws SSH2Exception, IOException, TimeoutException {

        final PlatformDescriptor platform = PlatformDescriptor.getPlatform(connection_wrapper);
        final String kill_command = getKillCommand(match, platform);

        try {
            runProcess(kill_command, connection_wrapper, null, null, null, null, KILL_TIMEOUT).waitFor();
        }
        catch (final InterruptedException e) {
            // Ignore.
        }
        catch (final NullPointerException e) {
            // Work round Mindterm bug.
        }
    }

    /**
     * Attempts to kill all local processes whose name includes the specified string.
     * 
     * @param match the string to match
     * @throws IOException if an error occurs when executing the kill command
     */
    public static void killMatchingProcesses(final String match) throws IOException {

        final PlatformDescriptor platform = PlatformDescriptor.getPlatform();
        final String kill_command = getKillCommand(match, platform);

        try {
            runProcess(kill_command).waitFor();
        }
        catch (final InterruptedException e) {
            // Ignore.
        }
    }

    /**
     * Runs the specified class in a new local OS process, using the class path of the currently executing program. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws UnknownPlatformException if the operating system of the local host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args) throws IOException, UnknownPlatformException {

        return runJavaProcess(clazz, args, System.out, System.err);
    }

    /**
     * Runs the specified class in a new local OS process, using the class path of the currently executing program. Output is written to the specified streams.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws UnknownPlatformException if the operating system of the local host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final OutputStream output_stream, final OutputStream error_stream) throws IOException, UnknownPlatformException {

        final String java_command = PlatformDescriptor.getPlatform().getJavaCommand(clazz, args);
        return runProcess(java_command, output_stream, error_stream);
    }

    /**
     * Runs the specified class in a new local OS process, using the specified class path. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param class_path the class path
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws UnknownPlatformException if the operating system of the local host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final ClassPath class_path) throws IOException, UnknownPlatformException {

        return runJavaProcess(clazz, args, class_path, System.out, System.err);
    }

    /**
     * Runs the specified class in a new local OS process, using the specified class path. Output is written to the specified streams.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param class_path the class path
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws UnknownPlatformException if the operating system of the local host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final ClassPath class_path, final OutputStream output_stream, final OutputStream error_stream) throws IOException, UnknownPlatformException {

        final String java_command = PlatformDescriptor.getPlatform().getJavaCommand(clazz, args, class_path, null);
        return runProcess(java_command, output_stream, error_stream);
    }

    // -------------------------------------------------------------------------------------------------------

    // Methods to run remote Java commands.

    /**
     * Runs the specified class in a new remote OS process, using the class path of the currently executing program.
     * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     *
     * @return a handle to the new OS process
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final SSH2ConnectionWrapper connection_wrapper) throws IOException, SSH2Exception, TimeoutException, UnknownPlatformException {

        return runJavaProcess(clazz, new ArrayList<String>(), connection_wrapper);
    }

    /**
     * Runs the specified class in a new remote OS process, using the class path of the currently executing program.
     * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final SSH2ConnectionWrapper connection_wrapper) throws IOException, SSH2Exception, TimeoutException, UnknownPlatformException {

        final PlatformDescriptor platform = PlatformDescriptor.getPlatform(connection_wrapper);
        final String java_command = platform.getJavaCommand(clazz, args);

        return runProcess(java_command, connection_wrapper);
    }

    /**
     * Runs the specified class in a new remote OS process, using the specified class path.
     * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param class_path the class path
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final SSH2ConnectionWrapper connection_wrapper, final ClassPath class_path) throws IOException, SSH2Exception, TimeoutException, UnknownPlatformException {

        return runJavaProcess(clazz, args, connection_wrapper, class_path, DEFAULT_SSH_CONNECTION_TIMEOUT);
    }

    /**
     * Runs the specified class in a new remote OS process, using the specified class path.
     * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path. Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param class_path the class path
     * @param timeout the timeout for completing the issue of the command (NB not for the remote process to complete) in milliseconds
     * @return a handle to the new OS process
     *
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final SSH2ConnectionWrapper connection_wrapper, final ClassPath class_path, final int timeout) throws IOException, SSH2Exception, TimeoutException, UnknownPlatformException {

        final PlatformDescriptor platform = PlatformDescriptor.getPlatform(connection_wrapper);

        final String java_command = platform.getJavaCommand(clazz, args, class_path, null);

        return runProcess(java_command, connection_wrapper, timeout);
    }

    /**
     * Runs the specified class in a new remote OS process, after first remotely installing the specified libraries, which make up the class path.
     * The libraries are installed using 'wget' ({@link "http://www.gnu.org/software/wget/"}). Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param lib_urls the libraries on which the class depends, specified with a URL for each jar file
     * @param create_timestamped_subdir true if a new timestamped sub-directory should be created within the install directory
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final File java_bin_path, final SSH2ConnectionWrapper connection_wrapper, final URL[] lib_urls, final boolean create_timestamped_subdir) throws IOException, SSH2Exception, TimeoutException,
                    UnknownPlatformException {

        return runJavaProcess(clazz, args, java_bin_path, connection_wrapper, lib_urls, create_timestamped_subdir, DEFAULT_SSH_CONNECTION_TIMEOUT);
    }

    /**
     * Runs the specified class in a new remote OS process, after first remotely installing the specified libraries, which make up the class path.
     * The libraries are installed using 'wget' ({@link "http://www.gnu.org/software/wget/"}). Output is written to local System.out and System.err.
     *
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param lib_urls the libraries on which the class depends, specified with a URL for each jar file
     * @param create_timestamped_subdir true if a new timestamped sub-directory should be created within the install directory
     * @param timeout the timeout for starting the remote process (NB not for the remote process to complete) in milliseconds
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
     * @throws TimeoutException if the remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the remote host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final File java_bin_path, final SSH2ConnectionWrapper connection_wrapper, final URL[] lib_urls, final boolean create_timestamped_subdir, final int timeout) throws IOException, SSH2Exception, TimeoutException,
                    UnknownPlatformException {

        final PlatformDescriptor platform = PlatformDescriptor.getPlatform(connection_wrapper);

        final File wget_path = new File(platform.getWgetPath());
        File lib_install_dir = new File(platform.getTempPath());

        if (create_timestamped_subdir) {
            lib_install_dir = new File(getTimestampedTempPath(lib_install_dir));
        }
        final ClassPath class_path = getClassPathForTempDir(lib_urls, lib_install_dir);

        final String install_libs_command = getInstallLibsCommand(lib_urls, lib_install_dir, wget_path);
        final String java_command = platform.getJavaCommand(clazz, args, class_path, lib_install_dir, java_bin_path);

        final String install_and_run_command = combineCommands(install_libs_command, java_command);
        System.out.println("command: " + install_and_run_command);

        return runProcess(install_and_run_command, connection_wrapper, timeout);
    }

    /**
     * Runs the specified class either locally, or remotely using a specified classpath, or remotely using library URLs, as specified by the host descriptor.
     * 
     * @param clazz the class to be executed
     * @param args command line arguments
     * @param host_descriptor a host descriptor
     * @return a handle to the new OS process
     * 
     * @throws IOException if an error occurs when reading output from the process
     * @throws SSH2Exception if an SSH connection to a remote machine cannot be established
     * @throws TimeoutException if a remote process cannot be instantiated within the timeout period
     * @throws UnknownPlatformException if the operating system of the host cannot be established
     */
    public static Process runJavaProcess(final Class<?> clazz, final List<String> args, final HostDescriptor host_descriptor, final File java_bin_path) throws IOException, SSH2Exception, TimeoutException, UnknownPlatformException {

        final SSH2ConnectionWrapper ssh_client_wrapper = host_descriptor.getSSHClientWrapper();
        final URL[] application_urls = host_descriptor.getApplicationUrls();
        final ClassPath class_path = host_descriptor.getClassPath();

        if (ssh_client_wrapper != null) {

            if (application_urls != null) {
                System.out.println("rjp1");
                return runJavaProcess(clazz, args, java_bin_path, ssh_client_wrapper, application_urls, true, NODE_INSTANTIATION_TIMEOUT);
            }
            System.out.println("rjp2");
            return runJavaProcess(clazz, args, ssh_client_wrapper, class_path, NODE_INSTANTIATION_TIMEOUT);
        }
        System.out.println("rjp3");
        return runJavaProcess(clazz, args);
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a thread that reads from one stream and writes to another.
     * 
     * @param command_stream the stream to be read from
     * @param print_stream the stream to be written to
     * @return the new thread
     */
    private static Thread makeReaderThread(final BufferedReader command_stream, final PrintStream print_stream) {

        return new Thread() {

            @Override
            public void run() {

                String s = null;
                try {
                    while ((s = command_stream.readLine()) != null) {
                        print_stream.println(s);
                    }
                }
                catch (final IOException e) {
                    /* Ignore read errors; they mean process is done. */
                }
                finally {
                    try {
                        command_stream.close();
                    }
                    catch (final IOException e) {
                    }
                }
            }
        };
    }

    /**
     * Creates an output stream that processes bytes using the specified processor and passes further output to the specified stream.
     *
     * @param stream OutputStream
     * @param processor IStreamProcessor
     * @return OutputStream
     */
    private static OutputStream makeInterceptStream(final OutputStream stream, final IStreamProcessor processor) {

        return new OutputStream() {

            @Override
            public void write(final int b) throws IOException {

                // Allow further processing if the processor says so, or if there is no processor.
                final boolean further_processing = processor == null || processor.processByte(b);

                if (further_processing && stream != null) {
                    stream.write(b);
                }
            }

            @Override
            public void close() throws IOException {

                // Do nothing.
            }

            @Override
            public void flush() throws IOException {

                stream.flush();
            }
        };
    }

    /**
     * Creates a shell command that echos the given timestamp to /dev/null then executes the given command.
     * The purpose is to label the command so that it can be identified in a subsequent process listing without having to know its PID.
     * 
     * @param command the command
     * @param timestamp the timestamp
     * @return the labelled command
     */
    private static String makeLabelledCommand(final String command, final String timestamp) {

        return "echo " + timestamp + " >/dev/null 2>&1; " + command;
    }

    /**
     * Creates an action that executes a given remote command.
     * 
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param command the command to be executed
     * @param output_stream the stream to which output from the process should be directed
     * @param error_stream the stream to which error output from the process should be directed
     * @return the new action
     */
    private static IActionWithResult makeRunCommandAction(final SSH2ConnectionWrapper connection_wrapper, final String command, final OutputStream output_stream, final OutputStream error_stream) {

        return new IActionWithResult() {

            @Override
            public Object performAction() {

                try {

                    // Create the remote console to use for command execution.
                    final SSH2ConsoleRemote console = new SSH2ConsoleRemote(connection_wrapper.getSSH2Client().getConnection());

                    // Run the command (returns a boolean indicating success, ignored here).
                    console.command(command, output_stream, error_stream, true); // Set TTY; it's OK if the streams are null.

                    return console;
                }
                catch (final SSH2Exception e) {
                    return e;
                }
                catch (final IOException e) {
                    return e;
                }
                catch (final RuntimeException e) {
                    return e;
                }
            }
        };
    }

    private static String getKillCommand(final String match, final PlatformDescriptor platform) {

        // Construct a shell command of this form:
        //
        // ps auxw | grep <match> | sed 's/^[a-z]*[ ]*\([0-9]*\).*/\1/' | tr '\n' ' ' | sed 's/^\(.*\)/kill -9 \1/' > /tmp/kill1; chmod +x /tmp/kill1; /tmp/kill1

        final String temp_dir_path = getTimestampedTempPath(new File(platform.getTempPath()));

        final StringBuilder string_builder = new StringBuilder();
        final PipedCommandBuilder command = new PipedCommandBuilder(string_builder);

        // List all processes.
        command.append("ps auxw");

        // Find those containing the specified string.
        command.append("grep " + match);

        // Get the PIDS, assuming each is preceded by a username and whitespace.
        command.append("sed 's/^[a-z]*[ ]*\\([0-9]*\\).*/\\1/'");

        // Join the PIDs onto a single line.
        command.append("tr '\\n' ' '");

        // Construct a command to kill them.
        command.append("sed 's/^\\(.*\\)/kill -9 \\1/' > " + temp_dir_path);

        string_builder.append("; chmod +x ");
        string_builder.append(temp_dir_path);
        string_builder.append("; ");
        string_builder.append(temp_dir_path);

        return string_builder.toString();
    }

    /**
     * Creates a new Process object that wraps the given remote console. In particular it allows the remote process to be destroyed.
     * Note that this method does not itself create a new remote process.
     * 
     * @param connection_wrapper an SSH connection wrapper for a remote machine
     * @param label a label that can be used to identify the remote process
     * @param console the remote SSH console
     * @return the new Process object
     */
    private static Process makeProcessForConsole(final SSH2ConnectionWrapper connection_wrapper, final String label, final SSH2ConsoleRemote console) {

        return new Process() {

            @Override
            public void destroy() {

                try {
                    killMatchingProcesses(label, connection_wrapper);
                }
                catch (final Exception e) {
                    Diagnostic.trace(DiagnosticLevel.RUN, "exception while trying to kill remote process: ", e.getMessage());
                }

                console.close();
            }

            @Override
            public int exitValue() {

                return waitFor();
            }

            @Override
            public InputStream getErrorStream() {

                return null;
            }

            @Override
            public InputStream getInputStream() {

                return console.getStdOut();
            }

            @Override
            public OutputStream getOutputStream() {

                return console.getStdIn();
            }

            @Override
            public int waitFor() {

                try {
                    return console.waitForExitStatus();
                }
                catch (final NullPointerException e) {
                    // Allow for Mindterm bug.
                    return -1;
                }
            }
        };
    }

    /**
     * Concatenates two shell commands.
     * 
     * @param command1 a command
     * @param command2 a command
     * @return a composite command that does the first then the second
     */
    private static String combineCommands(final String command1, final String command2) {

        if (command1.equals("")) { return command2; }
        if (command2.equals("")) { return command1; }

        return command1 + "; " + command2;
    }

    /**
     * Creates a shell command that downloads the specified files using wget.
     * 
     * @param lib_urls the URLs of the files to be downloaded
     * @param install_dir the directory to which the files should be downloaded
     * @param wget_path the path of the wget command
     * @return the composite command
     */
    private static String getInstallLibsCommand(final URL[] lib_urls, final File install_dir, final File wget_path) {

        String command = "";

        for (final URL lib_url : lib_urls) {
            final String install_lib_command = wget_path + " " + lib_url + " -q --directory-prefix=" + install_dir.getAbsolutePath();
            command = combineCommands(command, install_lib_command);
        }

        System.out.println("command: " + command);
        return command;
    }

    /**
     * Creates a path to a timestamped file or directory within the specified 'tmp' directory.
     * 
     * @param remote_temp_dir the 'tmp' directory
     * @return the path
     */
    private static String getTimestampedTempPath(final File remote_temp_dir) {

        return new File(remote_temp_dir, "tmp" + System.currentTimeMillis()).getAbsolutePath();
    }

    /**
     * Creates a classpath containing the specified library files, assuming that they have been downloaded to the specified directory.
     * 
     * @param lib_urls the URLs of the library files
     * @param download_dir the directory to which the files have been downloaded
     * @return the classpath
     */
    private static ClassPath getClassPathForTempDir(final URL[] lib_urls, final File download_dir) {

        final ClassPath class_path = new ClassPath();

        for (final URL url : lib_urls) {
            class_path.add(0, new File(download_dir, new File(url.getPath()).getName()));
        }

        return class_path;
    }
}
