/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.remote_management;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.Socket;

import uk.ac.standrews.cs.nds.util.MaskedStringInput;

import com.mindbright.jca.security.SecureRandom;
import com.mindbright.ssh2.SSH2Exception;
import com.mindbright.ssh2.SSH2SimpleClient;
import com.mindbright.ssh2.SSH2Transport;
import com.mindbright.util.RandomSeed;
import com.mindbright.util.SecureRandomAndPad;

/**
 * Wrapper for SSH2 connection details.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public final class SSH2ConnectionWrapper implements Cloneable {

    /**
     * Default port for SSH connection.
     */
    private static final int SSH_PORT = 22;

    private final InetAddress host;
    private final String user_name;

    private String password, key_file, key_passphrase;

    private SSH2SimpleClient client = null;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a new connection wrapper using a private key.
     * 
     * @param host the address of a remote host
     * @param user_name a user name
     * @param key_file the path of the user's private key file
     * @param key_passphrase the passphrase used to encrypt the private key file
     */
    public SSH2ConnectionWrapper(final InetAddress host, final String user_name, final String key_file, final String key_passphrase) {

        this.host = host;
        this.user_name = user_name;
        this.key_file = key_file;
        this.key_passphrase = key_passphrase;
    }

    /**
     * Creates a new connection wrapper using a password.
     * 
     * @param host the address of a remote host
     * @param user_name a user name
     * @param password the user's password
     */
    public SSH2ConnectionWrapper(final InetAddress host, final String user_name, final String password) {

        this.host = host;
        this.user_name = user_name;
        this.password = password;
    }

    // -------------------------------------------------------------------------------------------------------

    private SSH2ConnectionWrapper(final InetAddress host, final String user_name, final String password, final String key_file, final String key_passphrase) {

        this.host = host;
        this.user_name = user_name;
        this.password = password;
        this.key_file = key_file;
        this.key_passphrase = key_passphrase;
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Creates a connection wrapper using a host name and user credentials read from the user interactively.
     * The password or passphrase is input using a Swing dialog to prevent it being displayed on the screen.
     * 
     * @param use_password true if password authentication should be used, false if public key authentication should be used
     * @return the connection wrapper
     * @throws IOException if an error occurs while reading user input
     */
    public static SSH2ConnectionWrapper makeFromConsoleInput(final boolean use_password) throws IOException {

        final BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));

        System.out.print("enter machine name: ");
        final String machine_name = reader.readLine();

        final InetAddress address = InetAddress.getByName(machine_name);

        System.out.print("enter remote username for " + address.getCanonicalHostName() + ": ");
        final String user_name = reader.readLine();

        if (use_password) {

            final String password = MaskedStringInput.getMaskedString("enter remote password");

            return new SSH2ConnectionWrapper(address, user_name, password);
        }

        final String private_key_file_path = new File(new File(System.getProperty("user.home"), ".ssh"), "id_rsa").getAbsolutePath();
        final String pass_phrase = MaskedStringInput.getMaskedString("enter ssh passphrase");

        return new SSH2ConnectionWrapper(address, user_name, private_key_file_path, pass_phrase);
    }

    // -------------------------------------------------------------------------------------------------------

    @Override
    public SSH2ConnectionWrapper clone() {

        return new SSH2ConnectionWrapper(host, user_name, password, key_file, key_passphrase);
    }

    /**
     * Returns the remote host.
     * 
     * @return the remote host
     */
    public InetAddress getHost() {

        return host;
    }

    /**
     * Returns the user name.
     * 
     * @return the user name
     */
    public String getUserName() {

        return user_name;
    }

    /**
     * Returns the user's password.
     * 
     * @return the user's password
     */
    public String getPassword() {

        return password;
    }

    /**
     * Returns the path of the user's private key file.
     * 
     * @return the path of the user's private key file
     */
    public String getKeyFile() {

        return key_file;
    }

    /**
     * Returns the passphrase used to encrypt the private key file.
     * 
     * @return the passphrase used to encrypt the private key file
     */
    public String getKeyPassphrase() {

        return key_passphrase;
    }

    // -------------------------------------------------------------------------------------------------------

    protected SSH2SimpleClient getSSH2Client() throws SSH2Exception, IOException {

        if (client == null) {
            client = makeSSH2Client();
        }
        return client;
    }

    private SSH2SimpleClient makeSSH2Client() throws SSH2Exception, IOException {

        final Socket server_socket = new Socket(host, SSH_PORT);
        final byte[] seed = RandomSeed.getSystemStateHash();
        final SecureRandomAndPad secure_random = new SecureRandomAndPad(new SecureRandom(seed));

        final SSH2Transport transport = new SSH2Transport(server_socket, secure_random);

        if (password != null) { return new SSH2SimpleClient(transport, user_name, password); }

        return new SSH2SimpleClient(transport, user_name, key_file, key_passphrase);
    }
}
