/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.remote_management.test;

import java.net.URL;
import java.util.ArrayList;

import org.junit.Test;

import uk.ac.standrews.cs.nds.remote_management.PlatformDescriptor;
import uk.ac.standrews.cs.nds.remote_management.ProcessInvocation;
import uk.ac.standrews.cs.nds.remote_management.SSH2ConnectionWrapper;
import uk.ac.standrews.cs.nds.util.ClassPath;

/**
 * Various tests of local and remote process invocation, not intended to be run automatically.
 * They could be refactored to reduce code duplication, but it seems useful to keep them as self-contained examples.
 * 
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class ProcessInvocationTest {

    /**
     * Runs the 'uname' process locally.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runLocalProcess() throws Exception {

        final Process p = ProcessInvocation.runProcess("uname -a");

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a Java process locally.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runLocalJavaProcessShortLived() throws Exception {

        final Process p = ProcessInvocation.runJavaProcess(TestClassShortLived.class, new ArrayList<String>());

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a long-running Java process locally.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runLocalJavaProcessLongLived() throws Exception {

        ProcessInvocation.runJavaProcess(TestClassLongLived.class, new ArrayList<String>());
    }

    /**
     * Runs the 'uname' process remotely, using username/password authentication.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteProcessPassword() throws Exception {

        final String command = "uname -a";

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(true);
        final Process p = ProcessInvocation.runProcess(command, connection_wrapper);

        p.waitFor();
        p.destroy();
    }

    /**
     * Gets the platform of a remote machine.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void getRemotePlatform() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(true);
        final PlatformDescriptor platform = PlatformDescriptor.getPlatform(connection_wrapper);

        System.out.println("platform: " + platform);
    }

    /**
     * Runs the 'uname' process remotely, using username/public key authentication.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteProcessPublicKey() throws Exception {

        final String command = "uname -a";

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(false);
        final Process p = ProcessInvocation.runProcess(command, connection_wrapper);

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a Java process remotely, using username/password authentication.
     *
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteJavaProcessPassword() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(true);
        final Process p = ProcessInvocation.runJavaProcess(TestClassShortLived.class, new ArrayList<String>(), connection_wrapper, new ClassPath("/user/graham/nds.jar:/user/graham/stachord.jar"));

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a Java process remotely, using username/public key authentication.
     *
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteJavaProcessPublicKey() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(false);
        final Process p = ProcessInvocation.runJavaProcess(TestClassShortLived.class, new ArrayList<String>(), connection_wrapper, new ClassPath("/user/graham/nds.jar:/user/graham/stachord.jar"));

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a Java process remotely, using username/password authentication, and installing libraries dynamically.
     *
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteJavaProcessPasswordLibraryInstallation() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(true);
        final URL[] urls = new URL[]{new URL("http://www-systems.cs.st-andrews.ac.uk:8080/hudson/job/nds/lastStableBuild/artifact/bin/nds.jar")};
        final Process p = ProcessInvocation.runJavaProcess(TestClassShortLived.class, new ArrayList<String>(), null, connection_wrapper, urls, true);

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a Java process remotely, using username/public key authentication, and installing libraries dynamically.
     *
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteJavaProcessPublicKeyLibraryInstallation() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(false);
        final URL[] urls = new URL[]{new URL("http://www-systems.cs.st-andrews.ac.uk:8080/hudson/job/remote_management/lastStableBuild/artifact/bin/remote_management.jar")};
        final Process p = ProcessInvocation.runJavaProcess(TestClassShortLived.class, new ArrayList<String>(), null, connection_wrapper, urls, true);

        p.waitFor();
        p.destroy();
    }

    /**
     * Runs a long-running Java process remotely, using username/password authentication, and installing libraries dynamically.
     * This test should complete even though the remote process doesn't.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void runRemoteJavaProcessLongLived() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(true);
        final URL[] urls = new URL[]{new URL("http://www-systems.cs.st-andrews.ac.uk:8080/hudson/job/nds/lastStableBuild/artifact/bin/nds.jar")};

        ProcessInvocation.runJavaProcess(TestClassLongLived.class, new ArrayList<String>(), null, connection_wrapper, urls, true);
    }

    /**
     * Kills off a remote process using public key authentication.
     * 
     * @throws Exception if the test fails
     */
    @Test
    public void killRemoteProcessPublicKey() throws Exception {

        final SSH2ConnectionWrapper connection_wrapper = SSH2ConnectionWrapper.makeFromConsoleInput(false);

        ProcessInvocation.killMatchingProcesses("StartNodeInNewRing", connection_wrapper);
    }
}
