/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Queue of actions to be performed, serviced by a number of threads.
 * Queued actions may be performed in any order. Threads are created on
 * demand as actions arrive, up to a given limit. If a thread is idle
 * for more than a given period it dies off.
 * 
 * Now superseded by {@link Executors#newFixedThreadPool(int)}.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class ActionQueue {

    // -------------------------------------------------------------------------------------------------------

    /**
     * The default capacity of the queue.
     */
    public static final int DEFAULT_QUEUE_CAPACITY = 5;

    /**
     * The default maximum number of threads servicing the queue.
     */
    public static final int DEFAULT_MAX_THREADS = 2;

    /**
     * The default time-out in milliseconds after which an idle thread will die.
     */
    public static final int DEFAULT_IDLE_TIMEOUT = 5000;

    // -------------------------------------------------------------------------------------------------------

    private final List<IActionWithNoResult> action_list;
    private int number_of_threads;

    private final int max_threads, idle_timeout, queue_capacity;

    private final Semaphore mutex, not_empty, not_full;

    private final Object all_completed_sync = new Object();
    private int uncompleted_actions = 0;

    private static int thread_id = 0;

    // -------------------------------------------------------------------------------------------------------

    /**
     * Constructor with default capacity, number of service threads and idle time-out.
     */
    public ActionQueue() {

        this(DEFAULT_QUEUE_CAPACITY, DEFAULT_MAX_THREADS, DEFAULT_IDLE_TIMEOUT);
    }

    /**
     * Constructor with specified capacity and number of service threads.
     *
     * @param queue_capacity the capacity of the queue
     * @param max_threads the maximum number of service threads
     * @param idle_timeout the time-out in milliseconds after which an idle thread will die
     */
    public ActionQueue(final int queue_capacity, final int max_threads, final int idle_timeout) {

        this.queue_capacity = queue_capacity;
        this.max_threads = max_threads;
        this.idle_timeout = idle_timeout;

        // No threads are created initially.
        number_of_threads = 0;

        // Used for mutual exclusion on shared data structures.
        mutex = new Semaphore(1);

        // Used for synchronization on queue being empty.
        not_empty = new Semaphore(0);

        // Used for synchronization on queue being full.
        not_full = new Semaphore(queue_capacity);

        // Don't need a synchronized list (Collections.SynchronizedList()) since
        // all access to the ArrayList will be be synchronized.
        action_list = new ArrayList<IActionWithNoResult>();
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Queues the given action to be performed as soon as possible. Returns as soon as there is
     * space in the queue to store the action.
     *
     * @param action the action to be queued
     * @throws InterruptedException if the thread is interrupted before the action can be enqueued
     */
    public void enqueue(final IActionWithNoResult action) throws InterruptedException {

        // Wait until there is space in the queue.
        not_full.acquire();

        // Wait to enter critical section.
        mutex.acquire();

        // --------------- Start critical section ---------------

        // Add the action to the list.
        action_list.add(action);
        uncompleted_actions++;

        // If there are no idle threads, and the maximum number of threads hasn't been reached,
        // start a new one.
        if (not_empty.getQueueLength() == 0 && number_of_threads < max_threads) {

            newServiceThread();
            number_of_threads++;
        }

        // ---------------- End critical section ----------------

        // Leave critical section.
        mutex.release();

        // Signal availability of a new action.
        not_empty.release();
    }

    /**
     * Returns the number of free slots in the queue.
     *
     * @return the number of free slots
     * @throws InterruptedException if the thread is interrupted before the free space can be determined
     */
    public int freeSpace() throws InterruptedException {

        // Wait to enter critical section.
        mutex.acquire();

        // --------------- Start critical section ---------------

        // Calculate number of free slots.
        final int free_space = queue_capacity - action_list.size();

        // ---------------- End critical section ----------------

        // Leave critical section.
        mutex.release();

        return free_space;
    }

    /**
     * Blocks the caller until the queue is empty.
     * 
     * @throws InterruptedException if the thread is interrupted before the queue is empty
     */
    public void blockUntilNoUncompletedActions() throws InterruptedException {

        int uncompleted_actions = 0;

        // Wait to enter critical section.
        mutex.acquire();

        // --------------- Start critical section ---------------

        // Check whether queue is empty.
        uncompleted_actions = this.uncompleted_actions;

        // ---------------- End critical section ----------------

        // Leave critical section.
        mutex.release();

        // If the queue wasn't empty then block.
        synchronized (all_completed_sync) { // Need to synchronize to get ownership of the monitor.
            if (uncompleted_actions > 0) {
                all_completed_sync.wait();
            }
        }
    }

    // -------------------------------------------------------------------------------------------------------

    /**
     * Returns the next action in the queue. Delays until an action is available if the
     * queue is empty.
     *
     * @return the next action
     * @throws TimeoutException if the operation times out
     * @throws InterruptedException if the thread is interrupted before an action can be dequeued
     */
    private IActionWithNoResult dequeue() throws TimeoutException, InterruptedException {

        // Wait until there an action is available in the queue, or the timeout is exceeded.
        if (!not_empty.tryAcquire(idle_timeout, TimeUnit.MILLISECONDS)) {

            // No action available within timeout.
            throw new TimeoutException("ActionQueue::dequeue - timeout period exceeded");
        }

        // Wait to enter critical section.
        mutex.acquire();

        // --------------- Start critical section ---------------

        IActionWithNoResult next_action = null;

        // The queue could be empty even though 'not_empty' has been acquired, because another thread could
        // have emptied the queue while this thread was waiting to enter the critical section.
        if (!action_list.isEmpty()) {

            // Get the action at the front of the list.
            next_action = action_list.get(0);

            // Remove it from the list.
            action_list.remove(0);

            // Signal availability of space in the queue.
            not_full.release();
        }

        // ---------------- End critical section ----------------

        // Leave critical section.
        mutex.release();

        if (next_action != null) { return next_action; }

        // Slightly abusing semantics of InterruptedException, but the effect is the same for the application,
        // and in some sense this thread was interrupted from getting the next action from the queue.
        throw new InterruptedException();
    }

    /**
     * Creates and starts a new thread to service the action queue.
     */
    private void newServiceThread() {

        final Object queue = this;

        final Thread service_thread = new Thread("ActionQueue service thread " + ++thread_id) {

            @Override
            public void run() {

                try {
                    // Loop indefinitely removing actions from the queue.
                    while (true) {
                        dequeue().performAction();

                        synchronized (queue) {

                            uncompleted_actions--;

                            if (uncompleted_actions == 0) {

                                synchronized (all_completed_sync) { // Need to synchronize to get ownership of the monitor.
                                    all_completed_sync.notifyAll();
                                }
                            }
                        }
                    }
                }
                catch (final TimeoutException e) { // TimeoutException or InterruptedException

                    Diagnostic.trace(DiagnosticLevel.RUN, "ActionQueue service thread is terminating");

                    try {
                        mutex.acquire();

                        not_empty.release();
                        number_of_threads--;
                        mutex.release();
                    }
                    catch (final InterruptedException e1) {
                        ErrorHandling.exceptionError(e, "error letting thread die");
                    }

                    // Let thread die since it's been idle for too long.
                }
                catch (final InterruptedException e) {
                    // Ignore and try again next time loop.
                }
            }
        };

        service_thread.start();
    }
}
