/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.util.HashMap;
import java.util.Map;

/**
 * A global threshold diagnostic level may be set by the user.
 * The ordering of levels is FULL < INIT < RUN < RUNALL < RESULT < FINAL < NONE.
 *
 * @author Alan Dearle (al@cs.st-andrews.ac.uk)
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public enum DiagnosticLevel {

    /**
     * Level 0.
     */
    FULL(0),

    /**
     * Level 1.
     */
    INIT(1),

    /**
     * Level 2.
     */
    RUN(2),

    /**
     * Level 3.
     */
    RUNALL(3),

    /**
     * Level 4.
     */
    RESULT(4),

    /**
     * Level 5.
     */
    FINAL(5),

    /**
     * Level 6.
     */
    NONE(6);

    private static final Map<Integer, DiagnosticLevel> NUMERIC_MAP;

    private static DiagnosticLevel min_value;
    private static DiagnosticLevel max_value;

    static {

        NUMERIC_MAP = new HashMap<Integer, DiagnosticLevel>();

        for (final DiagnosticLevel level : DiagnosticLevel.values()) {
            NUMERIC_MAP.put(level.numericalValue(), level);

            if (min_value == null || level.numericalValue() < min_value.numericalValue()) {
                min_value = level;
            }
            if (max_value == null || level.numericalValue() > max_value.numericalValue()) {
                max_value = level;
            }
        }
    }

    private final int numerical_value;

    private DiagnosticLevel(final int numerical_value) {

        this.numerical_value = numerical_value;
    }

    /**
     * Returns the numerically lowest diagnostic level.
     * @return the numerically lowest diagnostic level
     */
    public static DiagnosticLevel getMin() {

        return min_value;
    }

    /**
     * Returns the numerically highest diagnostic level.
     * @return the numerically highest diagnostic level
     */
    public static DiagnosticLevel getMax() {

        return max_value;
    }

    /**
     * Returns the value of the numerically lowest diagnostic level.
     * @return the value of the numerically lowest diagnostic level
     */
    public static int getMinValue() {

        return getMin().numericalValue();
    }

    /**
     * Returns the value of the numerically highest diagnostic level.
     * @return the value of the numerically highest diagnostic level
     */
    public static int getMaxValue() {

        return getMax().numericalValue();
    }

    /**
     * Converts a numerical value to a diagnostic level.
     *
     * @param value a numerical value
     * @return the corresponding level
     * @throws UndefinedDiagnosticLevelException if the value does not have a corresponding level
     */
    public static DiagnosticLevel fromNumericalValue(final int value) throws UndefinedDiagnosticLevelException {

        if (NUMERIC_MAP.containsKey(value)) { return NUMERIC_MAP.get(value); }

        throw new UndefinedDiagnosticLevelException();
    }

    /**
     * Returns the numerical value of this diagnostic level.
     * @return the numerical value of this diagnostic level
     */
    public int numericalValue() {

        return numerical_value;
    }

    /**
     * Tests whether this diagnostic level meets a given threshold.
     *
     * @param threshold a threshold level
     * @return true if this diagnostic level is numerically greater than or equal to the value of the given level
     */
    public boolean meetsThreshold(final DiagnosticLevel threshold) {

        return numericalValue() >= threshold.numericalValue();
    }
}
