/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Contains various utility methods that make it easier to
 * print the contents of data structures such as collections, arrays, and maps.
 * 
 * @author Angus Macdonald (angus@cs.st-andrews.ac.uk)
 */
public final class PrettyPrinter {

    /**
    * Prevent instantiation of utility class.
    */
    private PrettyPrinter() {

    }

    /**
     * Concatenates representations of the full contents of a collection. Calls {@link #toString()} on each element.
     * 
     * @param collection a collection
     * @return a concatenated representation of the collection elements
     */
    public static String toString(final Collection<?> collection) {

        if (collection == null) { return null; }

        final StringBuilder sb = new StringBuilder();
        sb.append("Collection [");

        boolean first = true;
        for (final Object element : collection) {
            if (!first) {
                sb.append(", ");
            }
            sb.append(element);
            first = false;
        }

        sb.append("]");
        return sb.toString();
    }

    /**
     * Concatenates representations of the full contents of an array. Calls {@link #toString()} on each element.
     * 
     * @param array an array
     * @return a concatenated representation of the array elements
     */
    public static String toString(final Object[] array) {

        if (array == null) { return null; }

        final StringBuilder sb = new StringBuilder();
        sb.append("Array [");

        boolean first = true;
        for (final Object element : array) {
            if (!first) {
                sb.append(", ");
            }
            sb.append(element);
            first = false;
        }

        sb.append("]");

        return sb.toString();
    }

    /**
     * Concatenates representations of the full contents of a map. Calls {@link #toString()} on each element.
     * 
     * @param map a map
     * @return a concatenated representation of the map keys and values
     */
    public static String toString(final Map<?, ?> map) {

        if (map == null) { return null; }

        final StringBuilder sb = new StringBuilder();
        sb.append("Map [");

        boolean first = true;
        for (final Entry<?, ?> entry : map.entrySet()) {
            if (!first) {
                sb.append("\n");
            }
            sb.append("Key: " + entry.getKey() + "\t\tValue" + entry.getValue());
            first = false;
        }

        sb.append("]");

        return sb.toString();
    }

    /**
     * Returns the result of {@link #toString()} on the parameter.
     * 
     * This method doesn't add anything extra, but is a convenience if the parameter
     * to the PrettyPrint class is changed from a collection/map/array to a simple object.
     * 
     * @param object an object
     * @return the result of the {@link #toString()} method.
     */
    public static String toString(final Object object) {

        if (object == null) { return null; }

        return "Object [" + object.toString() + "]";
    }
}
