/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.util.concurrent.TimeoutException;

/**
 * Utility class to allow an action to be performed subject to a specified timeout.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class Timeout {

    /**
     * Performs an action subject to a specified timeout, in a separate thread. If a timeout occurs the action thread will continue to
     * execute.
     *
     * @param action an action to be performed
     * @param timeout a timeout in milliseconds
     * @throws TimeoutException if the timeout period elapses before the action has completed
     */
    public synchronized void performActionWithTimeout(final IActionWithNoResult action, final int timeout) throws TimeoutException {

        final Object sync = this;
        final boolean[] timed_out_flag_wrapper = new boolean[]{true};

        final Thread action_thread = new Thread() {

            @Override
            public void run() {

                try {
                    action.performAction();
                }
                catch (final RuntimeException e) {
                    timed_out_flag_wrapper[0] = false;
                    synchronized (sync) {
                        sync.notify();
                    }
                    throw e;
                }
                timed_out_flag_wrapper[0] = false;
                synchronized (sync) {
                    sync.notify();
                }
            }
        };
        action_thread.start();

        final Thread timeout_thread = new Thread() {

            @Override
            public void run() {

                try {
                    Thread.sleep(timeout);
                }
                catch (final InterruptedException e) { /* Ignore */
                }
                synchronized (sync) {
                    sync.notify();
                }
                action_thread.interrupt();
            }
        };
        timeout_thread.start();

        try {
            sync.wait();
        }
        catch (final InterruptedException e) { /* Ignore */
        }

        if (timed_out_flag_wrapper[0]) { throw new TimeoutException("timed out"); }
    }

    /**
     * Performs an action subject to a specified timeout, in a separate thread. If a timeout occurs the action thread will continue to
     * execute.
     *
     * @param action an action to be performed
     * @param timeout a timeout in milliseconds
     * @return the result of the action
     * @throws TimeoutException if the timeout period elapses before the action has completed
     */
    public synchronized Object performActionWithTimeout(final IActionWithResult action, final int timeout) throws TimeoutException {

        final Object[] result_wrapper = new Object[]{null};

        final IActionWithNoResult get_result = new IActionWithNoResult() {

            @Override
            public void performAction() {

                result_wrapper[0] = action.performAction();
            }
        };

        performActionWithTimeout(get_result, timeout);
        return result_wrapper[0];
    }
}
