/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util;

import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.util.Arrays;
import java.util.List;

/**
 * URI utility manipulation class.
 * 
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public final class UriUtil {

    /**
    * Prevent instantiation of utility class.
    */
    private UriUtil() {

    }

    /**
     * Replaces illegal URI characters with 3 char UTF-8 encoding.
     * 
     * @param s the string to be encoded
     * @return the corresponding legal URI string
     */
    public static String uriEncode(final String s) {

        try {
            String result = URLEncoder.encode(s, "UTF-8");

            // Above replaces spaces with + characters, and escapes /.

            // Replace + with %20.
            result = result.replaceAll("\\+", "%20");

            // Replace %2F with /.
            result = result.replaceAll("%2F", "/");

            return result;
        }
        catch (final UnsupportedEncodingException e) {
            ErrorHandling.hardExceptionError(e, "UTF-8 not supported");
            return null;
        }
    }

    /**
     * Creates a local path URI from a given parent and base name. This assumes an absolute path and ignores
     * scheme, host, port etc.
     * 
     * @param parent_uri the parent URI
     * @param base_name the base name
     * @param trailing_slash true if the new URI should include a trailing slash
     * @return the new URI
     */
    public static URI childUri(final URI parent_uri, final String base_name, final boolean trailing_slash) {

        String uri_path = parent_uri.toString();

        if (!uri_path.endsWith("/")) {
            uri_path += "/";
        }

        uri_path += uriEncode(base_name);

        if (trailing_slash) {
            uri_path += "/";
        }

        return stringToUri(uri_path);
    }

    /**
     * Creates a  local path URI corresponding to the parent of a given URI. This assumes an absolute path and ignores
     * scheme, host, port etc.
     * 
     * @param uri the child URI
     * @return the parent URI
     */
    public static URI parentUri(final URI uri) {

        final StringBuilder parent_path = new StringBuilder("/");

        final Iterable<String> element_iterator = pathElementIterator(uri);

        String previous = "";
        String current = "";

        for (final String element : element_iterator) {
            previous = current;
            current = element;
            if (parent_path.length() > 1) {
                parent_path.append("/");
            }
            parent_path.append(UriUtil.uriEncode(previous));
        }

        return stringToUri(parent_path.toString());
    }

    /**
     * Returns the base name of the path of the given URI.
     * 
     * @param uri the URI
     * @return the base name of the URI
     */
    public static String baseName(final URI uri) {

        String base_name = "";
        final Iterable<String> element_iterator = pathElementIterator(uri);

        for (final String element : element_iterator) {
            base_name = element;
        }

        return base_name;
    }

    /**
     * Returns an iterator over the names in the path of the given URI.
     * 
     * @param uri the URI
     * @return an iterator over the names in the path of the URI
     */
    public static Iterable<String> pathElementIterator(final URI uri) {

        String path = uri.normalize().getPath();

        if (path.startsWith("/")) {
            path = path.substring(1);
        }

        final String[] path_elements = path.split("/");

        List<String> path_as_list = Arrays.asList(path_elements);

        while (path_as_list.size() > 0 && path_as_list.get(0).equals("")) {
            path_as_list = path_as_list.subList(1, path_as_list.size());
        }

        return path_as_list;
    }

    /**
     * Tests whether the path of one resource is a strict ancestor of the path of another.
     * For example, "a/b" is considered an ancestor of "a/b/c" but not vice-versa.
     * 
     * @param resource1 the first resource
     * @param resource2 the second resource
     * @return true if the path of the first resource is an ancestor of the path of the second resource
     */
    public static boolean pathAncestorOf(final URI resource1, final URI resource2) {

        String path1 = resource1.normalize().getPath();
        if (!path1.endsWith("/")) {
            path1 += "/";
        }

        final String path2 = resource2.normalize().getPath();

        return !path1.equals(path2) && path2.startsWith(path1);
    }

    private static URI stringToUri(final String path) {

        try {
            return new URI(path);
        }
        catch (final URISyntaxException e) {
            ErrorHandling.exceptionError(e, "URI syntax error");
            return null;
        }
    }
}
