/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.math.BigInteger;

import org.junit.Test;

import uk.ac.standrews.cs.nds.p2p.impl.Key;

/**
 * Test class for {@link Key}.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class KeyTest {

    private static Key k1 = new Key(new BigInteger("-1"));
    private static Key k2 = new Key(BigInteger.ZERO);
    private static Key k3 = new Key(BigInteger.ONE);
    private static Key k4 = new Key(new BigInteger("3247823487234"));
    private static Key k5 = new Key(Key.KEYSPACE_SIZE.subtract(BigInteger.ONE));
    private static Key k6 = new Key(Key.KEYSPACE_SIZE);
    private static Key k7 = new Key(Key.KEYSPACE_SIZE.add(BigInteger.ONE));
    private static Key k8 = new Key(Key.KEYSPACE_SIZE.add(Key.KEYSPACE_SIZE.add(BigInteger.ONE)));
    private static Key k9 = new Key(new BigInteger("-1").multiply(Key.KEYSPACE_SIZE.add(Key.KEYSPACE_SIZE)).add(BigInteger.ONE));

    /**
     * Tests whether the key integers are as expected. All should lie in range zero to keyspace_size - 1.
     */
    @Test
    public void bigIntegerRepresentation() {

        // -1 should wrap to keyspace_size - 1.
        assertEquals(k1.keyValue(), Key.KEYSPACE_SIZE.subtract(BigInteger.ONE));

        // Original integers were within range, so should remain the same.
        assertEquals(k2.keyValue(), BigInteger.ZERO);
        assertEquals(k3.keyValue(), BigInteger.ONE);
        assertEquals(k4.keyValue(), new BigInteger("3247823487234"));
        assertEquals(k5.keyValue(), Key.KEYSPACE_SIZE.subtract(BigInteger.ONE));

        // keyspace_size should wrap to 0.
        assertEquals(k6.keyValue(), BigInteger.ZERO);

        // keyspace_size + 1 should wrap to 1.
        assertEquals(k7.keyValue(), BigInteger.ONE);

        // 2 * keyspace_size + 1 should wrap to 1.
        assertEquals(k8.keyValue(), BigInteger.ONE);

        // -2 * keyspace_size + 1 should wrap to 1.
        assertEquals(k9.keyValue(), BigInteger.ONE);
    }

    /**
     * Tests prefix matching.
     */
    @Test
    public void hexPrefixMatch() {

        // Check that zero and one have 39 hex digits in common.
        final int radix = 16;
        final int all_but_one_digit = 39;
        final int arbitrary_digits = 29;

        assertEquals(k2.countSharedPrefixDigits(k3, radix), all_but_one_digit);

        // Check that FFF..F and one have 0 digits in common.
        assertEquals(k1.countSharedPrefixDigits(k2, radix), 0);

        // Check that 2f4315d8102 and one have 1 digits in common.
        assertEquals(k4.countSharedPrefixDigits(k3, radix), arbitrary_digits);

        // Same checks in reverse.

        assertEquals(k3.countSharedPrefixDigits(k2, radix), all_but_one_digit);
        assertEquals(k2.countSharedPrefixDigits(k1, radix), 0);
        assertEquals(k3.countSharedPrefixDigits(k4, radix), arbitrary_digits);
    }

    /**
     * Tests whether the bit lengths of the keys are as expected.
     */
    @Test
    public void bitLength() {

        final int key_length = 160;
        final int arbitrary_length = 42;

        assertEquals(k1.bitLength(), key_length);
        assertEquals(k2.bitLength(), 0);
        assertEquals(k3.bitLength(), 1);
        assertEquals(k4.bitLength(), arbitrary_length);
        assertEquals(k5.bitLength(), key_length);
        assertEquals(k6.bitLength(), 0);
        assertEquals(k7.bitLength(), 1);
        assertEquals(k8.bitLength(), 1);
        assertEquals(k9.bitLength(), 1);
    }

    /**
     * Tests whether key comparison works as expected.
     */
    @Test
    public void comparison() {

        // k1 is the largest possible key, keyspace_size - 1.
        assertEquals(k1.compareTo(k2), 1);
        assertEquals(k1.compareTo(k3), 1);
        assertEquals(k1.compareTo(k4), 1);

        // k1 = k5.
        assertEquals(k1.compareTo(k5), 0);

        // k2 is the smallest possible key, zero.
        assertEquals(k2.compareTo(k1), -1);
        assertEquals(k2.compareTo(k3), -1);
        assertEquals(k2.compareTo(k4), -1);

        // k2 = k6.
        assertEquals(k2.compareTo(k6), 0);

        // Miscellaneous pairs.
        assertEquals(k4.compareTo(k1), -1);
        assertEquals(k4.compareTo(k2), 1);
        assertEquals(k4.compareTo(k3), 1);
        assertEquals(k4.compareTo(k4), 0);
    }

    /**
     * Tests whether ring distance calculation works as expected.
     */
    @Test
    public void ringDistanceTo() {

        // Distances clockwise from keyspace_size - 1.
        assertEquals(k1.ringDistanceTo(k2), BigInteger.ONE);
        assertEquals(k1.ringDistanceTo(k3), Key.TWO);
        assertEquals(k1.ringDistanceTo(k4), new BigInteger("3247823487235"));

        // k1 = k5.
        assertEquals(k1.ringDistanceTo(k5), BigInteger.ZERO);
        //assertEquals(k1.ringDistanceTo(k5), SegmentArithmetic.KEYSPACE_SIZE);

        // k2 is the smallest possible key, zero.
        assertEquals(k2.ringDistanceTo(k1), Key.KEYSPACE_SIZE.subtract(BigInteger.ONE));
        assertEquals(k2.ringDistanceTo(k3), BigInteger.ONE);
        assertEquals(k2.ringDistanceTo(k4), new BigInteger("3247823487234"));

        // k2 = k6.
        assertEquals(k2.ringDistanceTo(k6), BigInteger.ZERO);
        //assertEquals(k2.ringDistanceTo(k6), SegmentArithmetic.KEYSPACE_SIZE);

        // Miscellaneous pairs.
        assertEquals(k4.ringDistanceTo(k1), new BigInteger("1461501637330902918203684832716283016408109055741"));
        assertEquals(k4.ringDistanceTo(k2), new BigInteger("1461501637330902918203684832716283016408109055742"));
        assertEquals(k4.ringDistanceTo(k3), new BigInteger("1461501637330902918203684832716283016408109055743"));
        assertEquals(k4.ringDistanceTo(k4), BigInteger.ZERO);
        //assertEquals(k4.ringDistanceTo(k4), SegmentArithmetic.KEYSPACE_SIZE);
    }

    /**
     * Tests whether ring ordering calculation works as expected.
     */
    @Test
    public void ringOrdering() {

        assertTrue(k1.firstCloserInRingThanSecond(k2, k3));
        assertTrue(k1.firstCloserInRingThanSecond(k3, k4));
        assertTrue(k1.firstCloserInRingThanSecond(k2, k4));
        assertTrue(k1.firstCloserInRingThanSecond(k1, k3));

        assertFalse(k1.firstCloserInRingThanSecond(k3, k2));
        assertFalse(k1.firstCloserInRingThanSecond(k4, k3));
        assertFalse(k1.firstCloserInRingThanSecond(k4, k2));
        assertFalse(k1.firstCloserInRingThanSecond(k1, k1));
        assertFalse(k1.firstCloserInRingThanSecond(k2, k2));

        assertTrue(k2.firstCloserInRingThanSecond(k3, k4));
        assertTrue(k2.firstCloserInRingThanSecond(k4, k1));
        assertTrue(k2.firstCloserInRingThanSecond(k3, k1));
        assertTrue(k2.firstCloserInRingThanSecond(k2, k3));

        assertFalse(k2.firstCloserInRingThanSecond(k4, k3));
        assertFalse(k2.firstCloserInRingThanSecond(k1, k4));
        assertFalse(k2.firstCloserInRingThanSecond(k1, k3));
        assertFalse(k2.firstCloserInRingThanSecond(k2, k2));
        assertFalse(k2.firstCloserInRingThanSecond(k3, k3));

        assertTrue(k4.firstCloserInRingThanSecond(k1, k2));
        assertTrue(k4.firstCloserInRingThanSecond(k2, k3));
        assertTrue(k4.firstCloserInRingThanSecond(k1, k3));
        assertTrue(k4.firstCloserInRingThanSecond(k4, k3));

        assertFalse(k4.firstCloserInRingThanSecond(k2, k1));
        assertFalse(k4.firstCloserInRingThanSecond(k3, k2));
        assertFalse(k4.firstCloserInRingThanSecond(k3, k1));
        assertFalse(k4.firstCloserInRingThanSecond(k4, k4));
        assertFalse(k4.firstCloserInRingThanSecond(k3, k3));
    }
}
