/***************************************************************************
 *                                                                         *
 * nds Library                                                             *
 * Copyright (C) 2005-2010 Distributed Systems Architecture Research Group *
 * University of St Andrews, Scotland                                      *
 * http://www-systems.cs.st-andrews.ac.uk/                                 *
 *                                                                         *
 * This file is part of nds, a package of utility classes.                 *
 *                                                                         *
 * nds is free software: you can redistribute it and/or modify             *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * nds is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with nds.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                         *
 ***************************************************************************/
package uk.ac.standrews.cs.nds.util.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;

import org.junit.Test;

import uk.ac.standrews.cs.nds.util.UriUtil;

/**
 * Tests URI utilities.
 *
 * @author Graham Kirby (graham@cs.st-andrews.ac.uk)
 */
public class UriUtilTest {

    /**
     * Tests URI encoding.
     */
    @Test
    public void uriEncode() {

        assertEquals(UriUtil.uriEncode("my files"), "my%20files");
        assertEquals(UriUtil.uriEncode("a/b/c"), "a/b/c");
        assertEquals(UriUtil.uriEncode("a b/c d/e f"), "a%20b/c%20d/e%20f");
        assertEquals(UriUtil.uriEncode("/P2P Services.Data/rdb copy/"), "/P2P%20Services.Data/rdb%20copy/");
    }

    /**
     * Tests URI child processing.
     * 
     * @throws URISyntaxException if a test fails
     */
    @Test
    public void childUri() throws URISyntaxException {

        assertEquals(UriUtil.childUri(new URI(""), "abc", false), new URI("/abc"));
        assertEquals(UriUtil.childUri(new URI(""), "abc", true), new URI("/abc/"));
        assertEquals(UriUtil.childUri(new URI("/"), "abc", false), new URI("/abc"));
        assertEquals(UriUtil.childUri(new URI("/"), "abc", true), new URI("/abc/"));
        assertEquals(UriUtil.childUri(new URI("/x"), "abc", false), new URI("/x/abc"));
        assertEquals(UriUtil.childUri(new URI("/x"), "abc", true), new URI("/x/abc/"));
        assertEquals(UriUtil.childUri(new URI("/x/"), "abc", false), new URI("/x/abc"));
        assertEquals(UriUtil.childUri(new URI("/x/"), "abc", true), new URI("/x/abc/"));
        assertEquals(UriUtil.childUri(new URI("/x/"), "my files", true), new URI("/x/my%20files/"));
        assertEquals(UriUtil.childUri(new URI("/my%20files/"), "my file", true), new URI("/my%20files/my%20file/"));
        assertEquals(UriUtil.childUri(new URI("/a/b"), "stream closed", false), new URI("/a/b/stream%20closed"));
    }

    /**
     * Tests URI parent processing.
     * 
     * @throws URISyntaxException if a test fails
     */
    @Test
    public void parentUri() throws URISyntaxException {

        assertEquals(UriUtil.parentUri(new URI("")), new URI("/"));
        assertEquals(UriUtil.parentUri(new URI("/")), new URI("/"));
        assertEquals(UriUtil.parentUri(new URI("/abc")), new URI("/"));
        assertEquals(UriUtil.parentUri(new URI("/abc/")), new URI("/"));
        assertEquals(UriUtil.parentUri(new URI("/abc/def")), new URI("/abc"));
        assertEquals(UriUtil.parentUri(new URI("/abc/def%20ghi/j%20kl%23")), new URI("/abc/def%20ghi"));
    }

    /**
     * Tests URI base name processing.
     * 
     * @throws URISyntaxException if a test fails
     */
    @Test
    public void baseName() throws URISyntaxException {

        assertEquals(UriUtil.baseName(new URI("")), "");
        assertEquals(UriUtil.baseName(new URI("/")), "");
        assertEquals(UriUtil.baseName(new URI("/abc")), "abc");
        assertEquals(UriUtil.baseName(new URI("/abc/")), "abc");
        assertEquals(UriUtil.baseName(new URI("/abc/def")), "def");
        assertEquals(UriUtil.baseName(new URI("/abc/def%20ghi/j%20kl%23")), "j kl#");
    }

    /**
     * Tests URI path iteration.
     * 
     * @throws URISyntaxException if a test fails
     */
    @Test
    public void pathElementIterator() throws URISyntaxException {

        Iterator<String> iterator = UriUtil.pathElementIterator(new URI("")).iterator();
        assertFalse(iterator.hasNext());

        iterator = UriUtil.pathElementIterator(new URI("/")).iterator();
        assertFalse(iterator.hasNext());

        iterator = UriUtil.pathElementIterator(new URI("/abc")).iterator();
        assertEquals(iterator.next(), "abc");
        assertFalse(iterator.hasNext());

        iterator = UriUtil.pathElementIterator(new URI("/abc/")).iterator();
        assertEquals(iterator.next(), "abc");
        assertFalse(iterator.hasNext());

        iterator = UriUtil.pathElementIterator(new URI("/abc/def%20ghi/j%20kl%23")).iterator();
        assertEquals(iterator.next(), "abc");
        assertEquals(iterator.next(), "def ghi");
        assertEquals(iterator.next(), "j kl#");
        assertFalse(iterator.hasNext());
    }

    /**
     * Tests URI ancestor processing.
     * 
     * @throws URISyntaxException if a test fails
     */
    @Test
    public void pathAncestorOf() throws URISyntaxException {

        assertTrue(UriUtil.pathAncestorOf(new URI("a/b"), new URI("a/b/c")));
        assertTrue(UriUtil.pathAncestorOf(new URI("a/b/"), new URI("a/b/c")));
        assertFalse(UriUtil.pathAncestorOf(new URI("a/b"), new URI("a/b")));
        assertFalse(UriUtil.pathAncestorOf(new URI("a/b/"), new URI("a/b")));
        assertFalse(UriUtil.pathAncestorOf(new URI("a/b"), new URI("a/b/")));
        assertFalse(UriUtil.pathAncestorOf(new URI("a/b/c"), new URI("a/b/")));
        assertFalse(UriUtil.pathAncestorOf(new URI("a/b"), new URI("a/bc")));
    }
}
